/***********************************************************/
/*     T2Net_it.h  --  T2Net algorithm Interface           */
/*                                                         */
/*     Version 1.02                                        */
/***********************************************************/

#define MAX_DIMX   15
#define MAX_DIMY   30       

#define CLASSIC_MODE   0
#define PENTOMINO_MODE 1

#define MAX_INVENTORY 20

#define MAX_TXT 80

#define MAX_PLAYERS  8
#define NAME_LENGTH 32

#define MAX_LEVEL 20

typedef struct
  {
  unsigned char field[MAX_DIMX][MAX_DIMY];
  short int pc;
  short int next_pc;
  short int x;
  short int y;
  short int rot;
  short int nb_inventory;
  short int inventory[MAX_INVENTORY];
  short int level;
  short int late_nb;
  short int miss_nb;
  BOOL  player_alive[MAX_PLAYERS];
  BOOL  pending_msgs;
  BOOL  pending_actions;
  } T2NetEventReq;

#define ACTION_NOTHING        0
#define ACTION_MOVE_PIECE     1
#define ACTION_PLAY_INVENTORY 2
#define ACTION_MESSAGE        3

#define LINE_ADDED  20

#define EMPTY_BLOCK 0
#define TAKEN_BLOCK 1

#define INVENTORY_A 3
#define INVENTORY_B 4
#define INVENTORY_C 5
#define INVENTORY_G 6
#define INVENTORY_N 7
#define INVENTORY_O 8
#define INVENTORY_Q 9
#define INVENTORY_R 10
#define INVENTORY_S 11

#define MOVE_LEFT      1
#define MOVE_RIGHT     2
#define MOVE_ROT_LEFT  3
#define MOVE_ROT_RIGHT 4
#define MOVE_DOWN      5
#define MOVE_DROP      6

#define NIL_PLAYER -1

typedef struct
  {
  short int typemove;
  } MovePiece;

typedef struct
  {
  short int player;
  } PlayInventory;

typedef struct
  {
  char txt[MAX_TXT];
  } Message;

typedef struct
  {
  short int action;
  union
    {
    MovePiece      mp;
    PlayInventory  pi;
    Message        msg;
    } utype;
  } T2NetEventRsp;

#define INFO_GAME   1
#define INFO_MSG    2
#define INFO_ACTION 3

#define INFO_OK           0
#define INFO_ERROR_INFO   1
#define INFO_ERROR_PLAYER 2

typedef struct
  {
  short int player;
  } GamePlayerReq;

typedef struct
  {
  short int info;
  union
    {
    GamePlayerReq gp;
    } utype;
  } T2NetAlgoInfoReq;

typedef struct
  {
  short int player;
  char name[NAME_LENGTH];
  unsigned char field[MAX_DIMX][MAX_DIMY];
  } GamePlayerRsp;

#define MAX_MSG    15
#define MAX_ACTION 20

typedef struct
  {
  short int source_player;
  short int target_player;
  char  special_block;
  } t_action;

typedef struct
  {
  short int nb;
  short int first;
  char  messages[MAX_MSG][MAX_TXT];
  short int player_msg[MAX_MSG];
  } MessagePlayerRsp;

typedef struct
  {
  short int nb;
  short int first;
  t_action  action[MAX_ACTION];
  } ActionPlayerRsp;

typedef struct
  {
  short int info;
  union
    {
    GamePlayerRsp    gp;
    MessagePlayerRsp mp;
    ActionPlayerRsp  ap;
    } utype;
  } T2NetAlgoInfoRsp;

typedef struct
  {
  char team[NAME_LENGTH];
  short int list[MAX_PLAYERS];
  short int nb;
  } team_game_t;

typedef struct
  {
  short int dx;
  short int dy;
  BOOL  special_blocks;
  BOOL  pentomino;
  team_game_t teams[MAX_PLAYERS];
  short int nb_teams;
  short int num_player;
  short int num_team;
  FARPROC AlgoInfoReq;
  short int algo_delay;
  } T2NetInitInd;

typedef struct
  {
  LPSTR algoname;
  LPSTR realname;
  LPSTR email;
  LPSTR description;
  LPSTR version;
  LPSTR date;
  LPSTR interfaceversion;
  BOOL  special_blocks;
  BOOL  pentomino;
  } T2NetInfoRsp;         

typedef struct
  {
  short int kill_player_order[MAX_PLAYERS];
  short int kill_team_order[MAX_PLAYERS];
  short int players_score[MAX_PLAYERS];
  short int teams_score[MAX_PLAYERS];
  short int players_games[MAX_PLAYERS];
  short int nb_lines[MAX_PLAYERS];
  short int nb_lines_added[MAX_PLAYERS];
  } T2NetEndReq;

typedef struct
  {
  short int action;
  union
    {
    Message        msg;
    } utype;
  } T2NetEndRsp;

typedef struct        
 {
 short int x;
 short int y;
 }
 t_coord;

typedef struct
 {
 t_coord tab[5];
 short int offx;
 short int offy;
 short int lx;
 short int ly;
 short int starty;
 } t_piece;

typedef struct
 {
 short int nb;
 t_piece p[4];
 } t_pieces;

