#include <windows.h>
#include "memory.h"
#include "T2net_it.h"
#include "T2net_ct.h"


#define MAX_NB_WITHOUT_MOVE_PIECE 1


LPSTR AlgoName    = "Sample algo 2";
LPSTR RealName    = "XunK Entertainment Team";
LPSTR Email       = "XunK@mail.dotcom.fr";
LPSTR Description = "Second sample algorithm."
                    "Illustrates how to use the special blocks.";
LPSTR Version     = "1.00";


unsigned char playing_field[MAX_DIMX][MAX_DIMY];
short int x_size, y_size;

short int number_of_teams;
team_game_t teams[MAX_PLAYERS];
short int my_number,my_team_number;


//========================================================================
/*________________________________________________________
 |                                                        |
 | T2NetInfo                                              |
 |                                                        |
 | Transmits general informations about the algorithm     |
 | to the main program. These informations are displayed  |
 | and used by the main program.                          |
 |________________________________________________________|*/

void __declspec(dllexport)
T2NetInfo(T2NetInfoRsp *rsp)
{
  rsp->algoname         = AlgoName;
  rsp->realname         = RealName;
  rsp->email            = Email;
  rsp->description      = Description;
  rsp->version          = Version;
  rsp->date             = __DATE__;
  rsp->interfaceversion = interfaceversion;
  rsp->special_blocks   = TRUE;
  rsp->pentomino        = FALSE;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | T2NetInit                                              |
 |                                                        |
 | Used by the algorithm to store relevant informations   |
 | transmitted by the main program and describing         |
 | parameters of the game that is going to start.         |
 |________________________________________________________|*/

void  __declspec(dllexport)
T2NetInit(T2NetInitInd init)
{
  short int i;

  x_size = init.dx;
  y_size = init.dy;

  my_team_number = init.num_team;
  my_number      = init.num_player;

  number_of_teams = init.nb_teams;
  for (i=0;i<number_of_teams;i++)
    teams[i] = init.teams[i];
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | number_of_holes_in_whole_field                         |
 |                                                        |
 | Returns the number of "holes" in the playing field.    |
 |________________________________________________________|*/

short int number_of_holes_in_whole_field(void)
{
  short int x,y,number_of_holes,flag;

  number_of_holes = 0;
  for (x=0; x<x_size; x++)
    {
    flag = 0;
    for (y=0; y<y_size; y++)
      {
      if ((flag == 0) && (playing_field[x][y] != 0)) flag = 1;
      if ((flag == 1) && (playing_field[x][y] == 0)) number_of_holes++;
      }
    }
  return number_of_holes;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | full_line                                              |
 |                                                        |
 | Returns 1 if there is no hole in line y, else 0.       |
 |________________________________________________________|*/

short int full_line(short int y)
{
  short int x;

  for (x=0; x<x_size; x++)
    if (playing_field[x][y] == 0)
      return 0;

  return 1;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | remove_line                                            |
 |                                                        |
 | Removes line y from the playing field. The lines above |
 | line y are moved one step down in the playing field and|
 | the line at the top of the field is filled with 0.     |
 |________________________________________________________|*/

void remove_line(short int y_suppressed)
{
  short int x,y;

  for (y=y_suppressed; y>0; y--)
    for (x=0; x<x_size; x++)
      playing_field[x][y] = playing_field[x][y-1];

  for (x=0; x<x_size; x++)
    playing_field[x][0] = 0;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | test_piece                                             |
 |                                                        |
 | Returns 1 if it's possible to put the piece at         |
 | position (x,y) in the playing field, 0 otherwise.      |
 |________________________________________________________|*/

short int test_piece(T2NetEventReq *req,
                     short int rotation,
                     short int x,
                     short int y)
{
  t_piece *p;
  short int i;

  p = &pieces[CLASSIC_MODE][req->pc].p[rotation];

  if ((y+p->ly > y_size) || (x+p->lx > x_size))
    return 0;

  for (i=0; i<4; i++)
    if (playing_field[x+p->tab[i].x][y+p->tab[i].y] != 0)
      return 0;

  return 1;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | drop_piece_in_column_x                                 |
 |                                                        |
 | Returns 1 if it's possible to drop the piece in column |
 | x, 0 otherwise.                                        |
 |________________________________________________________|*/

short int drop_piece_in_column_x(T2NetEventReq *req,
                                 short int rotation,
                                 short int x,
                                 short int *y_reached,
                                 short int *number_of_lines_suppressed,
                                 short int *number_of_holes)
{
  short int i,y;
  t_piece *p;

  y = 0;
  while (test_piece(req, rotation, x, y))
    y++;
  if (y == 0)
    return 0;
  y--;

  p = &pieces[CLASSIC_MODE][req->pc].p[rotation];

  *y_reached = y-1+p->ly;

  for (i=0; i<4; i++)
    playing_field[x+p->tab[i].x][y+p->tab[i].y] = 1;

  *number_of_lines_suppressed = 0;
  y = 0;
  while (y<y_size)
    {
    if (full_line(y))
      {
      remove_line(y);
      (*number_of_lines_suppressed)++;
      }
    y++;
    }

  *number_of_holes = number_of_holes_in_whole_field();

  return 1;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | play_piece                                             |
 |                                                        |
 | Tests all the possibilities to play the current piece. |
 | Each possible "shape" of the piece (depending on the   |
 | rotation) is dropped in each possible column.          |
 | For each combination rotation/column, the consequences |
 | on the playing field are evaluated and the "best"      |
 | choice is selected.                                    |
 |                                                        |
 | For this simple sample algorithm, the following        |
 | condition is used to evaluate the "best" choice :      |
 | 1) choose the combination creating a maximum of full   |
 |    lines in the playing field.                         |
 | 2) if 1) is the same for several combinations, choose  |
 |    the combination that drops the piece as "deep" as   |
 |    possible in the field (close to the bottom).        |
 | 3) if 1) and 2) are the same for several combinations, |
 |    choose the combination that minimizes the number of |
 |    holes in the field.                                 |
 |________________________________________________________|*/

void play_piece(T2NetEventReq *req, T2NetEventRsp *rsp)
{
  short int rotation,x,y,n_holes,n_full_lines;
  short int chosen_rotation,chosen_x,chosen_n_holes;
  short int chosen_y,chosen_n_full_lines;
  BOOL new_choice;


  chosen_rotation     = 0;
  chosen_x            = 0;
  chosen_y            = -1;
  chosen_n_holes      = 255;
  chosen_n_full_lines = -1;


  for (rotation=0; rotation < pieces[CLASSIC_MODE][req->pc].nb; rotation++)
    for (x=0; x+pieces[CLASSIC_MODE][req->pc].p[rotation].lx <= x_size; x++)
      {
      memcpy(playing_field, req->field, MAX_DIMX*MAX_DIMY);

      if (drop_piece_in_column_x(req, rotation, x, &y, &n_full_lines, &n_holes))
        {
        new_choice = FALSE;

        if (n_full_lines>chosen_n_full_lines)
          new_choice = TRUE;
        else
          if (n_full_lines==chosen_n_full_lines)
            if (y>chosen_y)
              new_choice = TRUE;
            else
              if (y==chosen_y)
                if (n_holes<chosen_n_holes)
                  new_choice = TRUE;

        if (new_choice)
          {
          chosen_rotation     = rotation;
          chosen_x            = x;
          chosen_y            = y;
          chosen_n_holes      = n_holes;
          chosen_n_full_lines = n_full_lines;
          }
        }
      }

  rsp->action = ACTION_MOVE_PIECE;

  if (chosen_rotation != req->rot)
    rsp->utype.mp.typemove = MOVE_ROT_RIGHT;
  else
    if (chosen_x < req->x)
      rsp->utype.mp.typemove = MOVE_LEFT;
    else
      if (chosen_x > req->x)
        rsp->utype.mp.typemove = MOVE_RIGHT;
      else
        rsp->utype.mp.typemove = MOVE_DROP;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | look_for_opponent                                      |
 |                                                        |
 | Returns an opponent number.                            |
 | For this simple sample algorithm, the lowest "alive    |
 | opponent" number is chosen. Alive means that this      |
 | opponent has not already been eliminated from the      |
 | current game.                                          |
 |________________________________________________________|*/

short int look_for_opponent(T2NetEventReq *req)
{
  short int i,j;
  short int opponent_number;

  for (i=0; i<number_of_teams; i++)
    {
    if (i != my_team_number)
      for (j=0;j<teams[i].nb;j++)
        {
        opponent_number = teams[i].list[j];

        if (req->player_alive[opponent_number])
          return opponent_number;
        }
    }

  return NIL_PLAYER;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | use_special_block                                      |
 |                                                        |
 | Indicates how to use the current special block of the  |
 | inventory.                                             |
 | For this simple sample algorithm, special blocks       |
 | A,B,O,Q,R are used on an opponent selected by          |
 | look_for_opponent, special block S is deleted from the |
 | inventory (not used), and special blocks C,G,N are     |
 | used on the algorithm's field.                         |
 |________________________________________________________|*/

void use_special_block(T2NetEventReq *req, T2NetEventRsp *rsp)
{
  switch(req->inventory[req->nb_inventory-1])
    {
    case INVENTORY_A:
    case INVENTORY_B:
    case INVENTORY_O:
    case INVENTORY_Q:
    case INVENTORY_R:
      rsp->utype.pi.player = look_for_opponent(req);
      break;

    case INVENTORY_S:
      rsp->utype.pi.player = NIL_PLAYER;
      break;

    case INVENTORY_C:
    case INVENTORY_G:
    case INVENTORY_N:
      rsp->utype.pi.player = my_number;
      break;
    }

  rsp->action = ACTION_PLAY_INVENTORY;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | T2NetEvent                                             |
 |                                                        |
 | The main program calls regularly T2NetEvent during the |
 | game to transmit informations to the algorithm (field, |
 | inventory, current piece, next piece...) and the       |
 | algorithm answers with its actions (moving the piece   |
 | like this, using the special block in the inventory on |
 | this player...).                                       |
 |                                                        |
 | For this simple sample algorithm, the "strategy" is :  |
 | if the algorithm has not moved the current piece for   |
 | more than MAX_NB_WITHOUT_MOVE_PIECE consecutive calls  |
 | from T2net (because it was playing special blocks),    |
 | then it's time to move the current piece.              |
 | Else, if the inventory is not empty it uses the current|
 | special block (the one at the top of the inventory).   |
 | Of course, if the inventory is empty, the algorithm    |
 | always moves the current piece.                        |
 |________________________________________________________|*/

void __declspec(dllexport)
T2NetEvent(T2NetEventReq *req, T2NetEventRsp *rsp)
{
  if ( (req->nb_inventory != 0) &&
       (req->late_nb <= MAX_NB_WITHOUT_MOVE_PIECE) )
    use_special_block(req,rsp);
  else
    play_piece(req,rsp);
}
