#include <windows.h>
#include "memory.h"
#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "time.h"
#include "T2net_it.h"
#include "T2net_ct.h"
#include "resource.h"


#define STATE_NORMAL                  1
#define STATE_AFTER_BOMB_CHAT_MESSAGE 2


LPSTR AlgoName    = "Sample algo 4";
LPSTR RealName    = "XunK Entertainment Team";
LPSTR Email       = "XunK@mail.dotcom.fr";
LPSTR Description = "Fourth sample algorithm.\r\n"
                    "Illustrates :\r\n"
                    "-parameters for algorithms,\r\n"
                    "-reading of chat messages,\r\n"
                    "-use of T2NetEnd.";
LPSTR Version     = "1.00";


unsigned char playing_field[MAX_DIMX][MAX_DIMY];
short int x_size, y_size;
short int kind_of_game;

short int number_of_teams;
team_game_t teams[MAX_PLAYERS];
short int my_number,my_team_number;

short int state;
short int number_of_player_to_bomb;

FARPROC InfoFct;

HINSTANCE hInst;
short int MAX_NB_WITHOUT_MOVE_PIECE;
short int MIN_Y_DANGER_LIMIT;
short int BOMB_CHAT_MESSAGE_PERCENTAGE;


//========================================================================

BOOL APIENTRY DllMain(HANDLE hModule,
                      DWORD  ul_reason_for_call,
                      LPVOID lpReserved)
{
  hInst = hModule;

  return TRUE;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | T2NetInfo                                              |
 |                                                        |
 | Transmits general informations about the algorithm     |
 | to the main program. These informations are displayed  |
 | and used by the main program.                          |
 |________________________________________________________|*/

void __declspec(dllexport)
T2NetInfo(T2NetInfoRsp *rsp)
{
  rsp->algoname         = AlgoName;
  rsp->realname         = RealName;
  rsp->email            = Email;
  rsp->description      = Description;
  rsp->version          = Version;
  rsp->date             = __DATE__;
  rsp->interfaceversion = interfaceversion;
  rsp->special_blocks   = TRUE;
  rsp->pentomino        = TRUE;

  MAX_NB_WITHOUT_MOVE_PIECE    =  1;
  MIN_Y_DANGER_LIMIT           =  8;
  BOMB_CHAT_MESSAGE_PERCENTAGE = 50;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | algo_parameters and T2NetParam                         |
 |                                                        |
 | Used to display a dialog box and modify the algorithm  |
 | parameters without rebuilding the algorithm.           |
 |                                                        |
 | T2NetParam is called by T2net each time you click on   |
 | the Param button next to the algorithm name in the     |
 | "Local Players/Algorithms Selection" window.           |
 |________________________________________________________|*/

BOOL FAR PASCAL algo_parameters(HWND     hDlg,
                                unsigned message,
                                WORD     wParam,
                                LONG     lParam)
{
  switch (message)
    {
    case WM_INITDIALOG:
      SetDlgItemInt(hDlg, IDC_EDIT1, MAX_NB_WITHOUT_MOVE_PIECE, TRUE);
      SetDlgItemInt(hDlg, IDC_EDIT2, MIN_Y_DANGER_LIMIT, TRUE);
      SetDlgItemInt(hDlg, IDC_EDIT3, BOMB_CHAT_MESSAGE_PERCENTAGE, TRUE);
      return TRUE;

    case WM_COMMAND:
      switch (wParam)
        {
        case IDOK:
          MAX_NB_WITHOUT_MOVE_PIECE    = GetDlgItemInt(hDlg, IDC_EDIT1,NULL, TRUE);
          MIN_Y_DANGER_LIMIT           = GetDlgItemInt(hDlg, IDC_EDIT2,NULL, TRUE);
          BOMB_CHAT_MESSAGE_PERCENTAGE = GetDlgItemInt(hDlg, IDC_EDIT3,NULL, TRUE);
          EndDialog(hDlg, TRUE);
          break;
        case IDCANCEL:
          EndDialog(hDlg, TRUE);
          break;
        }
    break;
    }
  return FALSE;
}

//========================================================================

void __declspec(dllexport)
T2NetParam(void)
{
  DialogBox(hInst, MAKEINTRESOURCE(PARAMETERS), NULL, algo_parameters);
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | T2NetInit                                              |
 |                                                        |
 | Used by the algorithm to store relevant informations   |
 | transmitted by the main program and describing         |
 | parameters of the game that is going to start.         |
 |________________________________________________________|*/

void  __declspec(dllexport)
T2NetInit(T2NetInitInd init)
{
  short int i;

  x_size = init.dx;
  y_size = init.dy;

  if (init.pentomino)
    kind_of_game = PENTOMINO_MODE;
  else
    kind_of_game = CLASSIC_MODE;

  my_team_number = init.num_team;
  my_number      = init.num_player;

  number_of_teams = init.nb_teams;
  for (i=0;i<number_of_teams;i++)
    teams[i] = init.teams[i];

  InfoFct = init.AlgoInfoReq;

  state = STATE_NORMAL;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | number_of_holes_in_whole_field                         |
 |                                                        |
 | Returns the number of "holes" in the playing field.    |
 |________________________________________________________|*/

short int number_of_holes_in_whole_field(void)
{
  short int x,y,number_of_holes,flag;

  number_of_holes = 0;
  for (x=0; x<x_size; x++)
    {
    flag = 0;
    for (y=0; y<y_size; y++)
      {
      if ((flag == 0) && (playing_field[x][y] != 0)) flag = 1;
      if ((flag == 1) && (playing_field[x][y] == 0)) number_of_holes++;
      }
    }
  return number_of_holes;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | full_line                                              |
 |                                                        |
 | Returns 1 if there is no hole in line y, else 0.       |
 |________________________________________________________|*/

short int full_line(short int y)
{
  short int x;

  for (x=0; x<x_size; x++)
    if (playing_field[x][y] == 0)
      return 0;

  return 1;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | remove_line                                            |
 |                                                        |
 | Removes line y from the playing field. The lines above |
 | line y are moved one step down in the playing field and|
 | the line at the top of the field is filled with 0.     |
 |________________________________________________________|*/

void remove_line(short int y_suppressed)
{
  short int x,y;

  for (y=y_suppressed; y>0; y--)
    for (x=0; x<x_size; x++)
      playing_field[x][y] = playing_field[x][y-1];

  for (x=0; x<x_size; x++)
    playing_field[x][0] = 0;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | test_piece                                             |
 |                                                        |
 | Returns 1 if it's possible to put the piece at         |
 | position (x,y) in the playing field, 0 otherwise.      |
 |________________________________________________________|*/

short int test_piece(T2NetEventReq *req,
                     short int rotation,
                     short int x,
                     short int y)
{
  t_piece *p;
  short int i;

  p = &pieces[kind_of_game][req->pc].p[rotation];

  if ((y+p->ly > y_size) || (x+p->lx > x_size))
    return 0;

  for (i=0; i<4+kind_of_game; i++)
    if (playing_field[x+p->tab[i].x][y+p->tab[i].y] != 0)
      return 0;

  return 1;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | drop_piece_in_column_x                                 |
 |                                                        |
 | Returns 1 if it's possible to drop the piece in column |
 | x, 0 otherwise.                                        |
 |________________________________________________________|*/

short int drop_piece_in_column_x(T2NetEventReq *req,
                                 short int rotation,
                                 short int x,
                                 short int *y_reached,
                                 short int *number_of_lines_suppressed,
                                 short int *number_of_holes)
{
  short int i,y;
  t_piece *p;

  y = 0;
  while (test_piece(req, rotation, x, y))
    y++;
  if (y == 0)
    return 0;
  y--;

  p = &pieces[kind_of_game][req->pc].p[rotation];

  *y_reached = y-1+p->ly;

  for (i=0; i<4+kind_of_game; i++)
    playing_field[x+p->tab[i].x][y+p->tab[i].y] = 1;

  *number_of_lines_suppressed = 0;
  y = 0;
  while (y<y_size)
    {
    if (full_line(y))
      {
      remove_line(y);
      (*number_of_lines_suppressed)++;
      }
    y++;
    }

  *number_of_holes = number_of_holes_in_whole_field();

  return 1;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | play_piece                                             |
 |                                                        |
 | Tests all the possibilities to play the current piece. |
 | Each possible "shape" of the piece (depending on the   |
 | rotation) is dropped in each possible column.          |
 | For each combination rotation/column, the consequences |
 | on the playing field are evaluated and the "best"      |
 | choice is selected.                                    |
 |                                                        |
 | For this simple sample algorithm, the following        |
 | condition is used to evaluate the "best" choice :      |
 | 1) choose the combination creating a maximum of full   |
 |    lines in the playing field.                         |
 | 2) if 1) is the same for several combinations, choose  |
 |    the combination that drops the piece as "deep" as   |
 |    possible in the field (close to the bottom).        |
 | 3) if 1) and 2) are the same for several combinations, |
 |    choose the combination that minimizes the number of |
 |    holes in the field.                                 |
 |________________________________________________________|*/

void play_piece(T2NetEventReq *req, T2NetEventRsp *rsp)
{
  short int rotation,x,y,n_holes,n_full_lines;
  short int chosen_rotation,chosen_x,chosen_n_holes;
  short int chosen_y,chosen_n_full_lines;
  BOOL new_choice;


  chosen_rotation     = 0;
  chosen_x            = 0;
  chosen_y            = -1;
  chosen_n_holes      = 255;
  chosen_n_full_lines = -1;


  for (rotation=0; rotation < pieces[kind_of_game][req->pc].nb; rotation++)
    for (x=0; x+pieces[kind_of_game][req->pc].p[rotation].lx <= x_size; x++)
      {
      memcpy(playing_field, req->field, MAX_DIMX*MAX_DIMY);

      if (drop_piece_in_column_x(req, rotation, x, &y, &n_full_lines, &n_holes))
        {
        new_choice = FALSE;

        if (n_full_lines>chosen_n_full_lines)
          new_choice = TRUE;
        else
          if (n_full_lines==chosen_n_full_lines)
            if (y>chosen_y)
              new_choice = TRUE;
            else
              if (y==chosen_y)
                if (n_holes<chosen_n_holes)
                  new_choice = TRUE;

        if (new_choice)
          {
          chosen_rotation     = rotation;
          chosen_x            = x;
          chosen_y            = y;
          chosen_n_holes      = n_holes;
          chosen_n_full_lines = n_full_lines;
          }
        }
      }

  rsp->action = ACTION_MOVE_PIECE;

  if (chosen_rotation != req->rot)
    rsp->utype.mp.typemove = MOVE_ROT_RIGHT;
  else
    if (chosen_x < req->x)
      rsp->utype.mp.typemove = MOVE_LEFT;
    else
      if (chosen_x > req->x)
        rsp->utype.mp.typemove = MOVE_RIGHT;
      else
        rsp->utype.mp.typemove = MOVE_DROP;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | look_for_opponent                                      |
 |                                                        |
 | Returns an opponent number.                            |
 | For this simple sample algorithm, the lowest "alive    |
 | opponent" number is chosen. Alive means that this      |
 | opponent has not already been eliminated from the      |
 | current game.                                          |
 |________________________________________________________|*/

short int look_for_opponent(T2NetEventReq *req)
{
  short int i,j;
  short int opponent_number;

  for (i=0; i<number_of_teams; i++)
    {
    if (i != my_team_number)
      for (j=0;j<teams[i].nb;j++)
        {
        opponent_number = teams[i].list[j];

        if (req->player_alive[opponent_number])
          return opponent_number;
        }
    }

  return NIL_PLAYER;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | look_for_opponent_with_O_in_field                      |
 |                                                        |
 | Returns the lowest "alive opponent with a O in his     |
 | field" number, or NIL_PLAYER if no match.              |
 |________________________________________________________|*/

short int look_for_opponent_with_O_in_field(T2NetEventReq *req)
{
  short int i,j;
  short int opponent_number;
  T2NetAlgoInfoReq iareq;
  T2NetAlgoInfoRsp iarsp;
  short int x,y;

  for (i=0; i<number_of_teams; i++)
    {
    if (i != my_team_number)
      for (j=0;j<teams[i].nb;j++)
        {
        opponent_number = teams[i].list[j];

        if (req->player_alive[opponent_number])
          {
          iareq.utype.gp.player = opponent_number;
          iareq.info            = INFO_GAME;
          InfoFct(&iareq,&iarsp);

          for (x=0; x<x_size; x++)
            for (y=0; y<y_size; y++)
              if (iarsp.utype.gp.field[x][y] == INVENTORY_O)
                return opponent_number;
          }
        }
    }

  return NIL_PLAYER;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | use_special_block                                      |
 |                                                        |
 | Indicates how to use the current special block of the  |
 | inventory.                                             |
 | For this simple sample algorithm, special blocks       |
 | A,B,O,Q,R are used on an opponent selected by          |
 | look_for_opponent, special block S is deleted from the |
 | inventory (not used), and special blocks C,G,N are     |
 | used on the algorithm's field.                         |
 |________________________________________________________|*/

void use_special_block(T2NetEventReq *req, T2NetEventRsp *rsp)
{
  switch(req->inventory[req->nb_inventory-1])
    {
    case INVENTORY_A:
    case INVENTORY_B:
    case INVENTORY_O:
    case INVENTORY_Q:
    case INVENTORY_R:
      rsp->utype.pi.player = look_for_opponent(req);
      break;

    case INVENTORY_S:
      rsp->utype.pi.player = NIL_PLAYER;
      break;

    case INVENTORY_C:
    case INVENTORY_G:
    case INVENTORY_N:
      rsp->utype.pi.player = my_number;
      break;
    }

  rsp->action = ACTION_PLAY_INVENTORY;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | y_highest_block_in_field                               |
 |                                                        |
 | Returns the y value of the highest block in the field. |
 | Don't forget that this is in fact the minimum y value  |
 | because the y axis starts at the top of the field.     |
 |________________________________________________________|*/

short int y_highest_block_in_field(void)
{
  short int y_highest_block;
  short int x,y;

  y_highest_block = y_size;
  y               = 0;
  while( (y < y_size) && (y_highest_block == y_size) )
    {
    for (x=0; x<x_size; x++)
      if (playing_field[x][y] != 0)
        y_highest_block = y;

    y++;
    }

  return(y_highest_block);
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | send_bomb_chat_message                                 |
 |                                                        |
 | Prepares the message to send in the chat just before   |
 | playing a bomb (special block O) in an opponent's      |
 | field.                                                 |
 | This message is :                                      |
 | Ready to receive a bomb <name of the victim> ?         |
 |________________________________________________________|*/

void send_bomb_chat_message(T2NetEventReq *req, T2NetEventRsp *rsp)
{
  T2NetAlgoInfoReq iareq;
  T2NetAlgoInfoRsp iarsp;

  iareq.utype.gp.player = number_of_player_to_bomb;
  iareq.info            = INFO_GAME;
  InfoFct(&iareq,&iarsp);

  rsp->action = ACTION_MESSAGE;
  sprintf(rsp->utype.msg.txt,
          "Ready to receive a bomb %s ?",
          iarsp.utype.gp.name);
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | is_there_a_partner_requesting_help                     |
 |                                                        |
 | If a partner has sent a message in the chat with the   |
 | word "help" inside, this function returns the number of|
 | this partner (even if the message is : "I don't want   |
 | help"...). Else, the function returns NIL_PLAYER.      |
 |________________________________________________________|*/

short int is_there_a_partner_requesting_help(T2NetEventReq *req)
{
  short int partner;
  short int nb_pending_messages;
  short int num_message;
  char  message[MAX_TXT];
  short int i,j;
  T2NetAlgoInfoReq iareq;
  T2NetAlgoInfoRsp iarsp;

  if (req->pending_msgs)
    {
    iareq.info = INFO_MSG;
    InfoFct(&iareq,&iarsp);

    nb_pending_messages = iarsp.utype.mp.nb;
    num_message         = iarsp.utype.mp.first;
    for (i=0; i<nb_pending_messages; i++)
      {
      partner = NIL_PLAYER;
      for (j=0;j<teams[my_team_number].nb;j++)
        if (iarsp.utype.mp.player_msg[num_message] ==
            teams[my_team_number].list[j])
          {
          partner = teams[my_team_number].list[j];
          break;
          }

      if (partner != NIL_PLAYER)
        {
        strcpy(message,_strlwr(iarsp.utype.mp.messages[num_message]));
        if (strstr(message,"help") != NULL)
          return partner;
        }

      num_message++;
      if (num_message == MAX_MSG)
        num_message = 0;
      }
    }

  return NIL_PLAYER;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | T2NetEvent_1                                           |
 |                                                        |
 | Called when the algorithm is in the STATE_NORMAL state.|
 |                                                        |
 | For this sample algorithm, the "strategy" is :         |
 |                                                        |
 | -if the algorithm has not moved the current piece for  |
 | more than MAX_NB_WITHOUT_MOVE_PIECE consecutive calls  |
 | from T2net (because it was playing special blocks or   |
 | sending messages in the chat) or if the inventory is   |
 | empty, then it moves the current piece.                |
 | -else, if the current special block in inventory is :  |
 |   -G or N : if a partner requests help in the chat,    |
 |    the algorithm plays the G or N on this partner.     |
 |    Else, the algorithm uses the block on its own field |
 |    when the situation becomes dangerous.               |
 |    To determine if the situation becomes dangerous, the|
 |    algorithm compares the y of the highest block in    |
 |    its field with a defined limit (MIN_Y_DANGER_LIMIT).|
 |   -O : if the algorithm finds an opponent with an O    |
 |    block in its field it draws a random number to      |
 |    decide whether or not sending a message in the chat |
 |    to the designated opponent just before bombing him. |
 |    If so, the algorithm sends the message and sets     |
 |    state to STATE_AFTER_BOMB_CHAT_MESSAGE to remember  |
 |    that it must really bomb the designated player      |
 |    (number_of_player_to_bomb) next time it's called by |
 |    T2net. If not so, the algorithm plays the O block   |
 |    immediately on the designated player.               |
 |    If the algorithm doesn't find an opponent with an O |
 |    block in its field, it plays the O block like usual.|
 |   -any other : the algorithm plays the special block.  |
 |________________________________________________________|*/

void T2NetEvent_1(T2NetEventReq *req, T2NetEventRsp *rsp)
{
  short int partner_to_help;

  partner_to_help = is_there_a_partner_requesting_help(req);

  if ( (req->nb_inventory != 0) &&
       (req->late_nb <= MAX_NB_WITHOUT_MOVE_PIECE) )
    switch(req->inventory[req->nb_inventory-1])
      {
      case INVENTORY_G:
      case INVENTORY_N:
        if (partner_to_help != NIL_PLAYER)
          {
          rsp->action          = ACTION_PLAY_INVENTORY;
          rsp->utype.pi.player = partner_to_help;
          }
        else
          {
          memcpy(playing_field, req->field, MAX_DIMX*MAX_DIMY);
          if (y_highest_block_in_field() < MIN_Y_DANGER_LIMIT)
            use_special_block(req,rsp);
          else
            play_piece(req,rsp);
          }
        break;

      case INVENTORY_O:
        number_of_player_to_bomb = look_for_opponent_with_O_in_field(req);
        if (number_of_player_to_bomb != NIL_PLAYER)
          {
          srand((unsigned)time(NULL));
          if ((rand()%100) < BOMB_CHAT_MESSAGE_PERCENTAGE)
            {
            send_bomb_chat_message(req,rsp); 
            state = STATE_AFTER_BOMB_CHAT_MESSAGE;
            }
          else
            {
            rsp->action          = ACTION_PLAY_INVENTORY;
            rsp->utype.pi.player = number_of_player_to_bomb;
            }
          }
        else
          use_special_block(req,rsp);
        break;

      default:
        use_special_block(req,rsp);
      }
  else
    play_piece(req,rsp);
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | T2NetEvent_2                                           |
 |                                                        |
 | Called when the algorithm is in the                    |
 | STATE_AFTER_BOMB_CHAT_MESSAGE state and is going to    |
 | really bomb the designated player whose number has been|
 | stored before (number_of_player_to_bomb).              |
 | After that, the variable "state" is reset to its normal|
 | value (STATE_NORMAL).                                  |
 |________________________________________________________|*/

void T2NetEvent_2(T2NetEventReq *req, T2NetEventRsp *rsp)
{
  rsp->action          = ACTION_PLAY_INVENTORY;
  rsp->utype.pi.player = number_of_player_to_bomb;

  state = STATE_NORMAL;
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | T2NetEvent                                             |
 |                                                        |
 | The main program calls regularly T2NetEvent during the |
 | game to transmit informations to the algorithm (field, |
 | inventory, current piece, next piece...) and the       |
 | algorithm answers with its actions (moving the piece   |
 | like this, using the special block in the inventory on |
 | this player...).                                       |
 |                                                        |
 | For this sample algorithm, T2NetEvent uses the         |
 | variable "state" to remember if its last action was :  |
 | 1) doing nothing special,                              |
 | or                                                     |
 | 2) sending a message before bombing a player.          |
 |                                                        |
 | In the first case, the algorithm applies its normal    |
 | strategy and calls T2NetEvent_1 to do so.              |
 | In the second case, the algorithm must now really bomb |
 | the designated player and calls T2NetEvent_2 to do so. |
 |________________________________________________________|*/

void __declspec(dllexport)
T2NetEvent(T2NetEventReq *req, T2NetEventRsp *rsp)
{
  switch(state)
  {
  case STATE_NORMAL :
    T2NetEvent_1(req,rsp);
    break;

  case STATE_AFTER_BOMB_CHAT_MESSAGE :
    T2NetEvent_2(req,rsp);
    break;
  }
}

//========================================================================
/*________________________________________________________
 |                                                        |
 | T2NetEnd                                               |
 |                                                        |
 | The main program calls T2NetEnd at the end of each     |
 | game to transmit informations to the algorithm (who is |
 | the winner, what are the current scores, in which order|
 | were eliminated the players...)                        |
 |                                                        |
 | This sample algorithm uses these informations to       |
 | (sometimes) return a message to display in the chat.   |
 |________________________________________________________|*/

void __declspec(dllexport)
T2NetEnd(T2NetEndReq *req, T2NetEndRsp *rsp)
{
  srand((unsigned)time(NULL));
  if ((rand() % 2) != 0)
    rsp->action = ACTION_NOTHING;
  else
    {
    rsp->action = ACTION_MESSAGE;
    if (req->kill_team_order[0] == my_team_number)
      sprintf(rsp->utype.msg.txt,"Good game, Another One ?");
    else
      sprintf(rsp->utype.msg.txt,"Good play %s !!!", teams[req->kill_team_order[0]].team);
    }
}
